@@ -2,7 +2,8 @@ package Inline;
 
 use strict;
 require 5.006;
-$Inline::VERSION = '0.49';
+$Inline::VERSION = '0.51';
+$Inline::VERSION = eval $Inline::VERSION;
 
 use AutoLoader 'AUTOLOAD';
 use Inline::denter;
@@ -192,12 +193,11 @@ sub bind {
 # Process delayed objects that don't have source code yet.
 #==============================================================================
 # This code is an ugly hack because of the fact that you can't use an
-# INIT block at "run-time proper". So we kill the warning for 5.6+ users
-# and tell them to use a Inline->init() call if they run into problems. (rare)
-my $lexwarn = ($] >= 5.006) ? 'no warnings;' : '';
+# INIT block at "run-time proper". So we kill the warning and tell users
+# to use an Inline->init() call if they run into problems. (rare)
 
 eval <<END;
-$lexwarn
+no warnings;
 \$INIT = \$INIT; # Needed by Sarathy's patch.
 sub INIT {
     \$INIT++;
@@ -737,8 +737,11 @@ sub check_config_file {
            my $config = join '', <CONFIG>;
            close CONFIG;
 
-           croak M62_invalid_config_file(File::Spec->catfile($DIRECTORY,$configuration_file))
-             unless $config =~ /^version :/;
+           unless($config =~ /^version :/) {
+             warn "\$load_cfg sub: \$config: *${config}*\n";
+             croak M62_invalid_config_file(File::Spec->catfile($DIRECTORY,$configuration_file));
+           }
+
            if(UNTAINT) {
              warn "In Inline::check_config_file(): Blindly untainting Inline configuration file information.\n"
                unless $o->{CONFIG}{NO_UNTAINT_WARN};
@@ -804,15 +807,15 @@ sub create_config_file {
         my($v,$d,$f) = File::Spec->splitpath($inline);
         $f = "" if $f eq 'Inline.pm';
         $inline = File::Spec->catpath($v,$d,$f);
-        #my $INC = "-I$inline -I" .
-        #          join(" -I", grep {(-d File::Spec->catdir($_,"Inline") or
-        #                             -d File::Spec->catdir($_,"auto","Inline")
-	  #		            )} @INC);
-	#system "$perl $INC -MInline=_CONFIG_ -e1 $dir"
-        my @INC = map { "-I$_" }
+
+        # P::RD may be in a different PERL5LIB dir to Inline (as happens with cpan smokers).
+        # Therefore we need to grep for it - otherwise, if P::RD *is* in a different PERL5LIB
+        # directory the ensuing rebuilt @INC will not include that directory and attempts to use
+        # Inline::CPP (and perhaps other Inline modules) will fail because P::RD isn't found.
+        my @_inc = map { "-I$_" }
        ($inline,
-        grep {(-d File::Spec->catdir($_,"Inline") or -d File::Spec->catdir($_,"auto","Inline"))} @INC);
-       system $perl, @INC, "-MInline=_CONFIG_", "-e1", "$dir"
+        grep {(-d File::Spec->catdir($_,"Inline") or -d File::Spec->catdir($_,"auto","Inline") or -e File::Spec->catdir($_,"Parse/RecDescent.pm"))} @INC);
+       system $perl, @_inc, "-MInline=_CONFIG_", "-e1", "$dir"
 	  and croak M20_config_creation_failed($dir);
 	return;
     }
